"""Command-line interface options for the readme-ai application."""

from __future__ import annotations

from typing import Optional

import click

from readmeai.config.enums import BadgeOptions, ImageOptions, LLMOptions


def prompt_for_image(
    context: Optional[click.Context],
    parameter: Optional[click.Parameter],
    value: Optional[str],
) -> str:
    """Prompt the user for a custom image URL."""
    if value == ImageOptions.FILE.name:
        return click.prompt("Provide an image file path: ")
    elif value == ImageOptions.URL.name:
        return click.prompt("Provide an image URL: ")
    elif value in ImageOptions.__members__:
        return ImageOptions[value].value
    else:
        raise click.BadParameter(f"Invalid image URL provided: {value}")


align = click.option(
    "-a",
    "--align",
    type=click.Choice(["center", "left"], case_sensitive=False),
    default="center",
    help="Align the text in the README.md file's header to the left or center.",
)

api = click.option(
    "--api",
    type=click.Choice([opt.value for opt in LLMOptions], case_sensitive=False),
    default=None,
    help="""LLM service to use for generating the README.md file. The following options are currently supported:\n

    - OFFLINE (Generate the README.md file without making any API calls) \n
    - OPENAI (OpenAI GPT-3.5) \n
    - VERTEX (Google Vertex AI) \n
    """,
)

badges = click.option(
    "-b",
    "--badges",
    type=click.Choice(
        [opt.value for opt in BadgeOptions], case_sensitive=False
    ),
    default=BadgeOptions.DEFAULT.value,
    help="""\
        Badge icon style types to select from when generating README.md badges. The following options are currently available:\n
        - default \n
        - flat \n
        - flat-square \n
        - for-the-badge \n
        - plastic \n
        - skills \n
        - skills-light \n
        - social \n
        """,
)

badge_color = click.option(
    "--badge-color",
    type=str,
    default="0080ff",
    help="Custom color for the badge icon. Provide a valid color name or hex code.",
)

emojis = click.option(
    "-e",
    "--emojis",
    is_flag=True,
    default=False,
    help="This option adds emojis to the README.md file's header sections. For example, the default header for the 'Overview' section generates the markdown code as '## Overview'. Adding the --emojis flag generates the markdown code as '## 📍 Overview'.",
)

image = click.option(
    "-i",
    "--image",
    type=click.Choice(
        [opt.name for opt in ImageOptions], case_sensitive=False
    ),
    default=ImageOptions.BLUE.name,
    callback=prompt_for_image,
    show_choices=True,
    help="""\
        Project logo image displayed in the README file header. The following options are currently supported:\n

        Default image options:\n
        - BLACK \n
        - BLUE \n
        - CLOUD \n
        - GRADIENT \n
        - GREY \n
        - PURPLE \n

        Custom image options:\n
        - FILE (use a local file path to an image) \n
        - LLM (use LLM multi-modal capabilities to generate an image) \n
        - URL (use a custom image URL) \n
        """,
)

language = click.option(
    "-l",
    "--language",
    default="en",
    help="Language to use for generating the README.md file. Default is English (en).",
)

max_tokens = click.option(
    "--max-tokens",
    default=3899,
    type=int,
    help="Maximum number of tokens to generate for each section of the README.md file.",
)

model = click.option(
    "-m",
    "--model",
    default="gpt-3.5-turbo",
    help="GPT language model to use for generating various sections of the README.md file.",
)

output = click.option(
    "-o",
    "--output",
    default="readme-ai.md",
    help="Output file name for your README file. Default name is 'readme-ai.md'.",
)

repository = click.option(
    "-r",
    "--repository",
    required=True,
    help="Provide a remote repository URL (GitHub, GitLab, BitBucket), or a local directory path to your project.",
)

temperature = click.option(
    "-t",
    "--temperature",
    default=1.0,
    type=click.FloatRange(0.0, 2.0, clamp=True),
    help="Setting the model's temperature to a higher value will yield more creative content generated, while a lower value will generate more predictable content.",
)

template = click.option(
    "--template",
    type=str,
    help="README template file to use for generating the README.md file.",
)

tree_depth = click.option(
    "--tree-depth",
    default=5,
    type=int,
    help="Maximum depth of the directory tree thats included in the README.md file.",
)



"""Custom exceptions class for the readme-ai application."""

from __future__ import annotations


class ReadmeAiException(Exception):
    """Base exception for the readme-ai application."""

    ...


class GitCloneError(ReadmeAiException):
    """Could not clone repository."""

    def __init__(self, repository: str, *args):
        self.repository = repository
        super().__init__(f"Failed to clone repository: {repository}", *args)


class GitValidationError(ReadmeAiException):
    """Could not validate repository."""

    def __init__(self, repository: str, *args):
        self.repository = repository
        super().__init__(
            f"Failed to validate the provided repository: {repository}", *args
        )


class FileSystemError(ReadmeAiException):
    """Exceptions related to file system operations."""

    def __init__(self, message, path, *args):
        self.file_path = path
        super().__init__(f"{message}: {path}", *args)


class FileReadError(FileSystemError):
    """Could not read file."""

    ...


class FileWriteError(FileSystemError):
    """Could not write file."""

    ...


class ReadmeGeneratorError(ReadmeAiException):
    """Exceptions related to readme generation."""

    def __init__(self, traceback, *args):
        self.traceback = traceback
        super().__init__(f"Error generating readme: {traceback}", *args)


class UnsupportedServiceError(ReadmeAiException):
    """Exceptions related to the LLMHandler class."""

    def __init__(self, message, *args):
        super().__init__(
            f"Unsupported LLM API service provided: {message}", *args
        )



"""Data models and functions for configuring the readme-ai CLI tool."""

from importlib import resources
from pathlib import Path
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field, validator

from readmeai.config.validators import GitValidator
from readmeai.core.file_io import FileHandler
from readmeai.core.logger import Logger
from readmeai.exceptions import FileReadError

logger = Logger(__name__)


class FileSettings(BaseModel):
    """File paths used by the readme-ai CLI tool."""

    dependency_files: str = Field(
        description="Top programming language dependency file name."
    )
    ignore_files: str = Field(
        description="File, directory, and extension names to ignore."
    )
    language_names: str = Field(
        description="Programming language names mapped to extension names."
    )
    language_setup: str = Field(
        description="Programming language install, run, and test commands."
    )
    shields_icons: str = Field(
        description="List of shields.io icons used to build markdown badges."
    )
    skill_icons: str = Field(
        description="List of skill icons used to build markdown badges."
    )


class GitSettings(BaseModel):
    """User repository settings, sanitized and validated by Pydantic."""

    discussions: str = "https://github.com/eli64s/readme-ai/discussions"
    repository: Union[str, Path]
    full_name: Optional[str]
    host: Optional[str]
    host_domain: Optional[str]
    name: Optional[str]

    _validate_repository = validator("repository", pre=True, always=True)(
        GitValidator.validate_repository
    )
    _validate_full_name = validator("full_name", pre=True, always=True)(
        GitValidator.validate_full_name
    )
    _set_host = validator("host", pre=True, always=True)(GitValidator.set_host)
    _set_host_domain = validator("host_domain", pre=True, always=True)(
        GitValidator.set_host_domain
    )
    _set_name = validator("name", pre=True, always=True)(GitValidator.set_name)


class ModelSettings(BaseModel):
    """LLM API settings used for generating text for the README.md file."""

    api: str
    content: Optional[str]
    endpoint: Optional[str]
    encoding: Optional[str]
    model: Optional[str]
    offline: Optional[bool]
    temperature: Optional[float]
    tokens: Optional[int]
    tokens_max: Optional[int]
    rate_limit: Optional[int]


class MarkdownSettings(BaseModel):
    """Markdown template blocks for the README.md file."""

    align: str
    default: str
    badge_color: str
    badge_style: str
    badges_dependencies: str
    badge_logo: str
    badges_shields: str
    badges_skills: str
    contribute: str
    emojis: bool
    features: str
    header: str
    image: str
    modules: str
    modules_widget: str
    overview: str
    quickstart: str
    slogan: str
    tables: str
    toc: str
    tree: str
    tree_depth: int


class PromptSettings(BaseModel):
    """Prompt templates to generate text for the README.md file."""

    avatar: str = Field(
        description="Prompt to generate a project logo .png file."
    )
    features: str = Field(
        description="Prompt to generate a list of project features."
    )
    overview: str = Field(description="Prompt to generate a project overview.")
    slogan: str = Field(
        description="Prompt to generate a short project slogan or tagline."
    )
    summaries: str = Field(
        description="Prompt to generate a summary of a code file."
    )


class AppConfig(BaseModel):
    """Nested data model to store all configuration settings."""

    files: FileSettings
    git: GitSettings
    llm: ModelSettings
    md: MarkdownSettings
    prompts: PromptSettings


class AppConfigModel(BaseModel):
    """Pydantic model to store all configuration settings."""

    app: AppConfig

    class Config:
        """Pydantic model configuration."""

        validate_assignment = True


class ConfigHelper(BaseModel):
    """Helper class to load additional configuration files."""

    conf: AppConfigModel
    dependency_files: Dict[str, str] = {}
    ignore_files: Dict[str, List[str]] = {}
    language_names: Dict[str, str] = {}
    language_setup: Dict[str, List[str]] = {}

    class Config:
        allow_mutation = True

    def __init__(self, conf: AppConfigModel, **data):
        super().__init__(conf=conf, **data)
        self.load_helper_files()

    def load_helper_files(self):
        """Load helper configuration files."""
        handler = FileHandler()
        conf_path_list = [
            self.conf.app.files.dependency_files,
            self.conf.app.files.ignore_files,
            self.conf.app.files.language_names,
            self.conf.app.files.language_setup,
        ]

        for path in conf_path_list:
            conf_dict = _get_config_dict(handler, path)

            if "dependency_files" in conf_dict:
                self.dependency_files.update(conf_dict["dependency_files"])
            if "ignore_files" in conf_dict:
                self.ignore_files.update(conf_dict["ignore_files"])
            if "language_names" in conf_dict:
                self.language_names.update(conf_dict["language_names"])
            if "language_setup" in conf_dict:
                self.language_setup.update(conf_dict["language_setup"])


def _get_config_dict(handler: FileHandler, file_path: str) -> dict:
    """Get configuration dictionary from TOML file."""
    try:
        resource_path = resources.files("readmeai.settings") / file_path
    except TypeError as exc:
        logger.debug(f"Error using importlib.resources: {exc}")

        try:
            import pkg_resources

            resource_path = Path(
                pkg_resources.resource_filename(
                    "readmeai", f"settings/{file_path}"
                )
            ).resolve()
        except Exception as exc:
            raise FileReadError(
                "Error loading file via pkg_resources", file_path
            ) from exc

    if resource_path.exists() is False:
        raise FileReadError("File not found", resource_path) from None

    return handler.read(resource_path)


def load_config(path: str = "config.toml") -> AppConfig:
    """Load configuration constants from TOML file."""
    handler = FileHandler()
    conf_dict = _get_config_dict(handler, path)
    return AppConfigModel.parse_obj({"app": conf_dict}).app


def load_config_helper(conf: AppConfigModel) -> ConfigHelper:
    """Load multiple configuration helper TOML files."""
    return ConfigHelper(conf=conf)



#!/usr/bin/env python3

"""Main module for the README-AI CLI application."""

__package__ = "readmeai"

import asyncio
import tempfile
import traceback
from pathlib import Path
from typing import Optional

from readmeai.cli.options import prompt_for_image
from readmeai.config.enums import ImageOptions
from readmeai.config.settings import (
    AppConfig,
    AppConfigModel,
    ConfigHelper,
    GitSettings,
    load_config,
    load_config_helper,
)
from readmeai.core.file_io import FileHandler
from readmeai.core.logger import Logger
from readmeai.core.preprocess import process_repository
from readmeai.core.utils import setup_environment
from readmeai.exceptions import ReadmeGeneratorError
from readmeai.llms.registry import model_handler
from readmeai.markdown.builder import build_readme_md
from readmeai.services.git_utils import clone_repository

logger = Logger(__name__)


def readme_agent(
    align: Optional[str],
    api: str,
    badges: Optional[str],
    badge_color: Optional[str],
    emojis: Optional[bool],
    image: Optional[str],
    # language: Optional[str],
    max_tokens: Optional[int],
    model: Optional[str],
    output: Optional[str],
    repository: str,
    temperature: Optional[float],
    # template: Optional[str],
    tree_depth: Optional[int],
) -> None:
    """Main method of the readme-ai CLI application."""
    try:
        conf = load_config()
        conf_model = AppConfigModel(app=conf)
        conf_helper = load_config_helper(conf_model)
        conf.git = GitSettings(repository=repository)
        conf.md = conf.md.copy(
            update={
                "align": align,
                "badges": badges,
                "badge_color": badge_color,
                "emojis": emojis,
                "tree_depth": tree_depth,
                "image": image
                if image != ImageOptions.URL.name
                else prompt_for_image(None, None, image),
            }
        )
        conf.llm = conf.llm.copy(
            update={
                "api": api,
                "model": model,
                "temperature": temperature,
                "tokens_max": max_tokens,
            }
        )
        logger.info(f"Repository validation: {conf.git}")
        logger.info(f"LLM API validation: {conf.llm}")

        setup_environment(conf, api)
        asyncio.run(readme_generator(conf, conf_helper, output))

    except Exception as exc:
        raise ReadmeGeneratorError(exc, traceback.format_exc()) from exc


async def readme_generator(
    conf: AppConfig, conf_helper: ConfigHelper, output: str
) -> None:
    """Orchestrates the README file generation process."""
    try:
        with tempfile.TemporaryDirectory() as temp_dir:
            await clone_repository(conf.git.repository, temp_dir)
            (
                file_context,
                dependencies,
                raw_files,
                tree,
                dependency_dict,
            ) = process_repository(conf, conf_helper, temp_dir)
            logger.info(f"Dependency Files:\n{dependency_dict}")
            logger.info(f"Dependencies:\n{dependencies}")
            logger.info(f"Directory Tree:\n{tree}")

            async with model_handler(conf).use_api() as llm:
                responses = await llm.batch_request(dependencies, raw_files)
                (
                    summaries,
                    features,
                    overview,
                    slogan,
                ) = responses
                conf.md.features = conf.md.features.format(features)
                conf.md.overview = conf.md.overview.format(overview)
                conf.md.slogan = slogan

        markdown_content = build_readme_md(
            conf, conf_helper, dependencies, summaries, temp_dir
        )
        FileHandler().write(Path(output), markdown_content)

        logger.info(f"README file successfully generated @ {output}")
        logger.info(f"Share your README with us @ {conf.git.discussions}")

    except Exception as exc:
        raise ReadmeGeneratorError(traceback.format_exc()) from exc



"""Abstract base class for all LLM implementations."""

import asyncio
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from typing import Any, Dict, List, Tuple, Union

import httpx

from readmeai.config.settings import AppConfig
from readmeai.core.logger import Logger
from readmeai.llms.prompts import (
    get_prompt_context,
    set_additional_prompt_contexts,
    set_summary_prompt_context,
)
from readmeai.llms.tokens import update_max_tokens

logger = Logger(__name__)


class ModelHandler(ABC):
    """Abstract base class for LLM API implementations."""

    def __init__(self, config: AppConfig) -> None:
        """Initializes the GPT language model API handler."""
        self.config = config
        self.logger = Logger(__name__)
        self._http_client()
        self._llm_settings()

    @abstractmethod
    def _llm_settings(self):
        """Initializes basic attributes for the class."""
        ...

    @abstractmethod
    async def _handle_response(
        self,
        index: str,
        prompt: str,
        tokens: int,
        raw_files: List[Tuple[str, str]] = None,
    ) -> Tuple[str, str]:
        """Handles LLM API response and returns the generated text."""
        ...

    @asynccontextmanager
    async def use_api(self) -> None:
        """Context manager for HTTP client used by the LLM API."""
        try:
            yield self
        finally:
            await self.close()

    async def close(self) -> None:
        """Closes the HTTP client."""
        await self.http_client.aclose()

    def _http_client(self):
        """Configures the HTTP client for the class."""
        self.http_client = httpx.AsyncClient(
            http2=True,
            timeout=20,
            limits=httpx.Limits(
                max_keepalive_connections=100, max_connections=200
            ),
        )
        self.rate_limit_semaphore = asyncio.Semaphore(
            self.config.llm.rate_limit
        )

    async def batch_request(
        self,
        dependencies: List[str],
        raw_files: List[Tuple[str, str]],
    ) -> List[str]:
        """Generates text for the README.md file using LLM API."""
        if self.config.llm.offline is True:
            return await self._handle_response(
                index=None, prompt=None, tokens=None, raw_files=raw_files
            )

        prompts = await set_summary_prompt_context(
            self.config, dependencies, raw_files
        )
        file_summaries_resp = await self._batch_prompts(prompts)

        additional_prompts = await set_additional_prompt_contexts(
            self.config, dependencies, file_summaries_resp
        )
        additional_resp = await self._batch_prompts(additional_prompts)

        return file_summaries_resp + additional_resp

    async def _batch_prompts(
        self, prompts: List[Union[str, Tuple[str, str]]], batch_size: int = 10
    ):
        """Processes prompts in batches and returns the generated text."""
        responses = []

        for batch in self._generate_batches(prompts, batch_size):
            batch_responses = await asyncio.gather(
                *[self._process_batch(prompt) for prompt in batch],
                return_exceptions=True,
            )
            responses.extend(batch_responses)

        return responses

    def _generate_batches(self, items: List[Any], batch_size: int):
        """Generator to create batches from a list of items."""
        for i in range(0, len(items), batch_size):
            yield items[i : i + batch_size]

    async def _process_batch(self, prompt: Dict[str, Any]) -> str:
        """Processes a prompt and returns the generated text."""
        if prompt["type"] == "summaries":
            return await self._handle_response_code_summary(prompt["context"])
        else:
            formatted_prompt = get_prompt_context(
                self.config, prompt["type"], prompt["context"]
            )
            tokens = update_max_tokens(self.tokens, formatted_prompt)
            _, summary = await self._handle_response(
                prompt["type"], formatted_prompt, tokens
            )
            return summary

    async def _handle_response_code_summary(
        self, file_context: List[Tuple[str, str]]
    ) -> List[Tuple[str, str]]:
        """Generates code summaries for the README.md file."""
        summary_text = []
        for context in file_context["summaries"]:
            file_path, file_content = context
            prompt = self.prompts.summaries.format(
                self.config.md.tree, file_path, file_content
            )
            tokens = update_max_tokens(self.tokens, prompt)
            _, summary_or_error = await self._handle_response(
                file_path, prompt, tokens
            )
            summary_text.append((file_path, summary_or_error))

        return summary_text



"""Enable offline mode when no LLM API service is specified."""

from typing import List, Tuple

from readmeai.config.settings import AppConfig
from readmeai.core.logger import Logger
from readmeai.core.model import ModelHandler

logger = Logger(__name__)


class OfflineModeHandler(ModelHandler):
    """Offline model handler."""

    def __init__(self, config: AppConfig) -> None:
        """Initialize OpenAI API LLM."""
        self.config = config
        self._llm_settings()
        super().__init__(config)

    def _llm_settings(self):
        """Set default values for offline mode."""
        self.placeholder = self.config.md.default
        self.prompts = self.config.prompts

    async def _handle_response(
        self,
        index: str = None,
        prompt: str = None,
        tokens: int = None,
        raw_files: List[Tuple[str, str]] = None,
    ) -> Tuple[str, str]:
        """Use default placeholders when offline mode is enabled."""
        file_summaries = [
            (str(file_path), self.placeholder) for file_path, _ in raw_files
        ]
        return (
            file_summaries,
            self.placeholder,
            self.placeholder,
            self.placeholder,
        )



"""Prompt manager for LLM API calls."""

from typing import List

import readmeai.config.settings as AppConfig
from readmeai.core.logger import Logger

logger = Logger(__name__)


def get_prompt_context(
    config: AppConfig, prompt_type: str, context: dict
) -> str:
    """Generates a prompt for the LLM API."""
    prompt_template = get_prompt_template(config.prompts, prompt_type)

    if not prompt_template:
        logger.error(f"Prompt type '{prompt_type}' not found.")
        return ""

    return inject_prompt_context(prompt_template, context)


def get_prompt_template(prompts: dict, prompt_type: str) -> str:
    """Retrieves the template for the given prompt type."""
    prompt_templates = {
        "features": prompts.features,
        "overview": prompts.overview,
        "slogan": prompts.slogan,
    }
    return prompt_templates.get(prompt_type, "")


def inject_prompt_context(template: str, context: dict) -> str:
    """Formats the template with the provided context."""
    try:
        return template.format(*[context[key] for key in context])
    except KeyError as exc:
        logger.error(f"Missing context for prompt key: {exc}")
        return ""


async def set_summary_prompt_context(
    config: AppConfig,
    dependencies: List[str],
    summaries: List[str],
) -> List[dict]:
    """Generates the summary prompts to be used by the LLM API."""
    return [
        {"type": prompt_type, "context": context}
        for prompt_type, context in [
            (
                "summaries",
                {
                    "tree": config.md.tree,
                    "dependencies": dependencies,
                    "summaries": summaries,
                },
            ),
        ]
    ]


async def set_additional_prompt_contexts(
    config: AppConfig,
    dependencies: List[str],
    summaries: List[str],
) -> List[dict]:
    """Generates additional prompts (features, overview, slogan) for the LLM API."""
    return [
        {"type": prompt_type, "context": context}
        for prompt_type, context in [
            (
                "features",
                {
                    "repo": config.git.repository,
                    "dependencies": dependencies,
                    "summaries": summaries,
                },
            ),
            (
                "overview",
                {
                    "name": config.git.name,
                    "repo": config.git.repository,
                    "summaries": summaries,
                },
            ),
            (
                "slogan",
                {
                    "name": config.git.name,
                    "repo": config.git.repository,
                    "summaries": summaries,
                },
            ),
        ]
    ]



"""Abstract factory for all dependency file parsers."""

from typing import Dict

from readmeai.core.parser import FileParser
from readmeai.parsers.cpp import (
    CMakeParser,
    ConfigureAcParser,
    MakefileAmParser,
)
from readmeai.parsers.docker import DockerComposeParser, DockerfileParser
from readmeai.parsers.go import GoModParser
from readmeai.parsers.gradle import BuildGradleKtsParser, BuildGradleParser
from readmeai.parsers.maven import MavenParser
from readmeai.parsers.npm import PackageJsonParser, YarnLockParser
from readmeai.parsers.python import RequirementsParser, TomlParser, YamlParser
from readmeai.parsers.rust import CargoTomlParser
from readmeai.parsers.swift import SwiftPackageParser


def parser_factory() -> Dict[str, FileParser]:
    """Returns a dictionary of callable file parser methods."""
    yaml_parser = YamlParser()
    toml_parser = TomlParser()
    requirements_parser = RequirementsParser()

    return {
        # Python
        "Pipfile": toml_parser,
        "pyproject.toml": toml_parser,
        "requirements.in": requirements_parser,
        "requirements.txt": requirements_parser,
        "requirements-dev.txt": requirements_parser,
        "requirements-test.txt": requirements_parser,
        "requirements-prod.txt": requirements_parser,
        "dev-requirements.txt": requirements_parser,
        "environment.yml": yaml_parser,
        "environment.yaml": yaml_parser,
        # "setup.py": setup_py_parser,
        # "setup.cfg": setup_cfg_parser,
        # C/C++
        "cmakeLists.txt": CMakeParser(),
        "configure.ac": ConfigureAcParser(),
        "Makefile.am": MakefileAmParser(),
        # JavaScript/Node.js
        "package.json": PackageJsonParser(),
        "yarn.lock": YarnLockParser(),
        # Kotlin and Kotlin DSL
        "build.gradle": BuildGradleParser(),
        "build.gradle.kts": BuildGradleKtsParser(),
        # Go
        "go.mod": GoModParser(),
        # Java
        "pom.xml": MavenParser(),
        # Rust
        "cargo.toml": CargoTomlParser(),
        # Swift
        "Package.swift": SwiftPackageParser(),
        "Dockerfile": DockerfileParser(),
        "docker-compose.yaml": DockerComposeParser(),
    }



"""CLI entrypoint for the readme-ai application."""

from typing import Optional

import click

from readmeai.app import readme_agent
from readmeai.cli import options


@click.command()
@options.align
@options.api
@options.badges
@options.badge_color
@options.emojis
@options.image
@options.language
@options.max_tokens
@options.model
@options.output
@options.repository
@options.temperature
@options.tree_depth
@options.template
def main(
    align: Optional[str],
    api: str,
    badges: Optional[str],
    badge_color: Optional[str],
    emojis: Optional[bool],
    image: Optional[str],
    language: Optional[str],
    max_tokens: Optional[int],
    model: Optional[str],
    output: Optional[str],
    repository: str,
    temperature: Optional[float],
    template: Optional[str],
    tree_depth: Optional[int],
) -> None:
    """Entry point for the readme-ai CLI application."""
    readme_agent(
        align=align,
        api=api,
        badges=badges,
        badge_color=badge_color,
        emojis=emojis,
        image=image,
        # language=language,
        max_tokens=max_tokens,
        model=model,
        output=output,
        repository=repository,
        temperature=temperature,
        # template=template,
        tree_depth=tree_depth,
    )


if __name__ == "__main__":
    main()



"""Processes the input repository files and extracts metadata."""

from dataclasses import dataclass, field
from pathlib import Path
from typing import Generator, List, Tuple

from readmeai.config.settings import AppConfig, ConfigHelper
from readmeai.core.logger import Logger
from readmeai.core.utils import is_file_ignored
from readmeai.llms.tokens import count_tokens
from readmeai.markdown.builder import ReadmeBuilder
from readmeai.parsers.registry import parser_factory

logger = Logger(__name__)

GITHUB_WORKFLOWS_PATH = ".github/workflows"
PARSERS = parser_factory()


@dataclass
class FileContext:
    """Data class to store file contents and metadata."""

    file_path: Path
    file_name: str
    file_ext: str
    content: str
    tokens: int = 0
    language: str = field(init=False)
    dependencies: List[str] = field(default_factory=list)

    def __post_init__(self):
        """Initializes the FileContext class."""
        self.file_ext = (
            self.file_name.split(".")[-1] if "." in self.file_name else ""
        )
        self.language = self.file_ext.lower()


class RepoProcessor:
    """Processes the repository files and generates FileContext list"""

    def __init__(
        self,
        config: AppConfig,
        config_helper: ConfigHelper,
    ):
        """Initializes the RepoProcessor class."""
        self.config = config
        self.config_helper = config_helper
        self.dependency_files = self.config_helper.dependency_files.get(
            "dependency_files", []
        )
        self.language_names = config_helper.language_names
        self.language_setup = config_helper.language_setup

    def create_file_data(
        self, file_info: Tuple[str, Path, str]
    ) -> FileContext:
        """Creates a FileContext instance from the file information."""
        file_name, file_path, content = file_info
        return FileContext(
            file_name=file_name,
            file_path=file_path,
            content=content,
            file_ext="",
        )

    def extract_dependencies(self, file_data: FileContext) -> List[str]:
        """Extracts the dependency file contents using the factory pattern."""
        parsers = PARSERS

        if file_data.file_name not in parsers:
            return []

        parser = parsers.get(file_data.file_name)
        dependency_names = parser.parse(content=file_data.content)
        logger.info(
            f"Dependency file found: {file_data.file_name}:\n{dependency_names}"
        )
        return dependency_names

    def generate_contents(self, repo_path: str) -> List[FileContext]:
        """Generates a List of Dict of file information."""
        if isinstance(repo_path, str):
            repo_path = Path(repo_path)

        return [file_data for file_data in self.generate_file_info(repo_path)]

    def generate_file_info(
        self, repo_path: Path
    ) -> Generator[FileContext, None, None]:
        """
        Generates FileContext instances for each file in the repository.
        """
        for file_path in repo_path.rglob("*"):
            if self._is_file_ignored(file_path):
                continue
            file_data = self._process_file_path(file_path, repo_path)
            if file_data:
                yield file_data

    def get_dependencies(self, contents: List[FileContext]) -> List[str]:
        """Returns a list of dependencies."""
        try:
            dependency_dict = {}
            dependencies = set()
            dependency_files = self.config_helper.dependency_files.get(
                "dependency_files"
            )

            for file_data in contents:
                dependencies.update(file_data.dependencies)
                dependencies.add(file_data.language)
                dependencies.add(file_data.file_ext)

                if file_data.file_name in dependency_files:
                    dependencies.add(file_data.file_name)
                    dependency_dict[
                        file_data.file_name
                    ] = file_data.dependencies

                if GITHUB_WORKFLOWS_PATH in str(file_data.file_path):
                    dependencies.add("github actions")

            return list(dependencies), dependency_dict

        except Exception as exc:
            logger.error(f"Error getting dependencies: {exc}")
            return []

    def language_mapper(
        self, contents: List[FileContext]
    ) -> List[FileContext]:
        """Maps file extensions to their programming languages."""
        for content in contents:
            content.language = self.language_names.get(
                content.file_ext, ""
            ).lower()
        return contents

    def tokenize_content(
        self, contents: List[FileContext]
    ) -> List[FileContext]:
        """Tokenize each file content and return the token count."""
        if self.config.llm.offline is True:
            return contents

        for content in contents:
            content.tokens = count_tokens(
                content.content, self.config.llm.encoding
            )
        return contents

    def _is_file_ignored(self, file_path: Path) -> bool:
        """
        Determines if a file should be ignored based on configurations.
        """
        if (
            is_file_ignored(self.config_helper, file_path)
            and str(file_path.name) in self.dependency_files
        ):
            return False

        return not file_path.is_file() or is_file_ignored(
            self.config_helper, file_path
        )

    def _process_file_path(
        self, file_path: Path, repo_path: Path
    ) -> FileContext:
        """
        Processes an individual file path and returns FileContext.
        """
        relative_path = file_path.relative_to(repo_path)
        if GITHUB_WORKFLOWS_PATH in str(relative_path):
            return FileContext(
                file_path=relative_path,
                file_name="github actions",
                file_ext="",
                content="",
            )

        try:
            with file_path.open(encoding="utf-8") as file:
                content = file.read()
            file_data = FileContext(
                file_path=relative_path,
                file_name=file_path.name,
                file_ext="",
                content=content,
            )
            file_data.dependencies = self.extract_dependencies(file_data)
            return file_data

        except (OSError, UnicodeDecodeError) as exc:
            logger.warning(f"Error reading file {file_path}: {exc}")


def process_repository(
    config: AppConfig, config_helper: ConfigHelper, temp_dir: str
) -> Tuple[List[FileContext], List[str], List[Tuple[str, str]], str]:
    """Processes the repository files and returns the context."""
    repo_processor = RepoProcessor(config, config_helper)
    repo_context = repo_processor.generate_contents(temp_dir)
    repo_context = repo_processor.tokenize_content(repo_context)
    repo_context = repo_processor.language_mapper(repo_context)

    dependencies, dependency_dict = repo_processor.get_dependencies(
        repo_context
    )

    raw_files = [
        (str(context.file_path), context.content) for context in repo_context
    ]

    config.md.tree = ReadmeBuilder(
        config, config_helper, dependencies, raw_files, temp_dir
    ).md_tree

    return (
        repo_context,
        dependencies,
        raw_files,
        config.md.tree,
        dependency_dict,
    )



"""Google Vertex AI LLM API implementation."""

import os
from typing import List, Tuple

import httpx
import vertexai
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from vertexai.preview.generative_models import GenerativeModel

from readmeai.config.settings import AppConfig
from readmeai.core.logger import Logger
from readmeai.core.model import ModelHandler
from readmeai.core.utils import format_response
from readmeai.llms.tokens import count_tokens, truncate_tokens

logger = Logger(__name__)


class VertexAIHandler(ModelHandler):
    """Vertex AI Generative Models API LLM implementation."""

    def __init__(self, config: AppConfig) -> None:
        """Initialize Vertex AI Generative Models API LLM."""
        self.config = config
        self._llm_settings()
        super().__init__(config)
        vertexai.init(project=self.project_id, location=self.location)

    def _llm_settings(self):
        """Initializes basic attributes for the class."""
        self.location = os.getenv("VERTEXAI_LOCATION", None)
        self.project_id = os.getenv("VERTEXAI_PROJECT", None)

        self.content = self.config.llm.content
        self.encoder = self.config.llm.encoding
        self.model = self.config.llm.model
        self.prompts = self.config.prompts
        self.tokens = self.config.llm.tokens
        self.tokens_max = self.config.llm.tokens_max
        self.temperature = self.config.llm.temperature

    @retry(
        stop=stop_after_attempt(4),
        wait=wait_exponential(multiplier=1, min=2, max=6),
        retry=retry_if_exception_type(
            (
                httpx.HTTPStatusError,
                httpx.NetworkError,
                httpx.TimeoutException,
            )
        ),
    )
    async def _handle_response(
        self,
        index: str,
        prompt: str,
        tokens: int,
        raw_files: List[Tuple[str, str]] = None,
    ) -> Tuple[str, str]:
        """Handles Vertex AI LLM API responses and returns the generated text."""
        token_cnt = count_tokens(prompt, self.encoder)
        if token_cnt > self.tokens_max:
            self.logger.debug(
                f"Truncating '{index}' prompt: {token_cnt} > {self.tokens_max}"
            )
            prompt = truncate_tokens(self.encoder, prompt, tokens)

        try:
            async with self.rate_limit_semaphore:
                model = GenerativeModel(self.model)
                response = await model.generate_content_async(
                    prompt,
                )
                response_text = response.text
                formatted_text = format_response(index, response_text)
                self.logger.info(f"Response for '{index}':\n{formatted_text}")
                return index, formatted_text

        except (
            httpx.HTTPStatusError,
            httpx.NetworkError,
            httpx.TimeoutException,
        ) as exc:
            logger.error(f"Error generating response for {index}: {exc}")
            return index, self.config.md.default



"""Functions for building and formatting the badges in the README.md file."""

from typing import Dict, Tuple

from readmeai.config.enums import BadgeOptions
from readmeai.config.settings import AppConfig
from readmeai.core.file_io import FileHandler
from readmeai.core.utils import get_resource_path
from readmeai.services.git_utils import GitService


def _read_badge_file(file_path: str) -> Dict[str, str]:
    """Reads the badges file and returns the SVG icons."""
    resource_path = get_resource_path(__package__, file_path)
    return FileHandler().read(resource_path)


def build_dependency_badges(
    dependencies: list[str], icons: dict[str, str], style: str
) -> str:
    """Build HTML badges for project dependencies."""
    badges = [
        icons[str(dependency).lower()]
        for dependency in dependencies
        if str(dependency).lower() in icons
    ]

    # Sort badges by hex value (from light to dark color)
    badges.sort(key=lambda b: int(b[1], 16) if b[1] else 0, reverse=True)
    badges = [badge[0].format(style) for badge in badges]
    return format_badges(badges)


def build_metadata_badges(config: AppConfig, full_name: str, host: str) -> str:
    """Build metadata badges using shields.io."""
    return config.md.badges_shields.format(
        host=host,
        full_name=full_name,
        badge_style=config.md.badge_style,
        badge_color=config.md.badge_color,
    )


def format_badges(badges: list[str]) -> str:
    """Format SVG badge icons as HTML."""
    len_icons = len(badges)
    if badges is None or len_icons == 0:
        return ""

    badges_per_line = (
        len_icons if len_icons < 9 else (len_icons // 2) + (len_icons % 2)
    )

    lines = []
    for i in range(0, len_icons, badges_per_line):
        line = "\n\t".join(
            [
                f'<img src="{badge}" alt="{badge.split("/badge/")[1].split("-")[0]}">'
                for badge in badges[i : i + badges_per_line]
            ]
        )
        lines.append(
            f"{line}\n\t<br>"
            if i + badges_per_line < len_icons
            else f"{line}\n"
        )

    return "\n\t".join(lines)


def shields_icons(
    conf: AppConfig, deps: list, full_name: str, git_host: str
) -> Tuple[str, str]:
    """
    Generates badges for the README using shields.io icons, referencing
    the repository - https://github.com/Aveek-Saha/GitHub-Profile-Badges.
    """
    badge_set = _read_badge_file(conf.files.shields_icons)

    metadata_badges = build_metadata_badges(conf, full_name, git_host)

    dependency_badges = build_dependency_badges(
        deps, badge_set, conf.md.badge_style
    )
    dependency_badges = conf.md.badges_dependencies.format(
        align=conf.md.align, badges=dependency_badges
    )

    if (
        conf.md.badge_style == BadgeOptions.DEFAULT.value
        and git_host != GitService.LOCAL
    ):
        return (
            metadata_badges,
            "<!-- default option, no dependency badges. -->\n",
        )

    if git_host == GitService.LOCAL:
        return (
            "<!-- local repository, no metadata badges. -->\n",
            dependency_badges,
        )

    return metadata_badges, dependency_badges


def skill_icons(conf: AppConfig, deps: list) -> str:
    """
    Generates badges for the README using skill icons, from the
    repository - https://github.com/tandpfun/skill-icons.
    """
    deps.extend(["md"])
    icons_dict = _read_badge_file(conf.files.skill_icons)
    icons_list = [
        icon for icon in icons_dict["icons"]["names"] if icon in deps
    ]
    skill_icons = ",".join(icons_list)
    # per_line = (len(skill_icons) + 2) // 2
    # icon_names = f"{icon_names}"  # &perline={per_line}"
    skill_icons = icons_dict["url"]["base_url"] + skill_icons

    if conf.md.badge_style == "skills-light":
        skill_icons = f"{skill_icons}&theme=light"

    conf.md.badges_skills = conf.md.badges_skills.format(skill_icons)
    dependency_badges = conf.md.badges_dependencies.format(
        align=conf.md.align, badges=conf.md.badges_skills
    )
    return dependency_badges



"""OpenAI API LLM implementation."""

from typing import List, Tuple

import httpx
import openai
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from readmeai.config.settings import AppConfig
from readmeai.core.logger import Logger
from readmeai.core.model import ModelHandler
from readmeai.core.utils import format_response
from readmeai.llms.tokens import count_tokens, truncate_tokens

logger = Logger(__name__)


class OpenAIHandler(ModelHandler):
    """OpenAI API LLM implementation."""

    def __init__(self, config: AppConfig) -> None:
        """Initialize OpenAI API LLM."""
        self.config = config
        self._llm_settings()
        super().__init__(config)

    def _llm_settings(self):
        self.content = self.config.llm.content
        self.encoder = self.config.llm.encoding
        self.endpoint = self.config.llm.endpoint
        self.model = self.config.llm.model
        self.prompts = self.config.prompts
        self.tokens = self.config.llm.tokens
        self.tokens_max = self.config.llm.tokens_max
        self.temperature = self.config.llm.temperature

    @retry(
        stop=stop_after_attempt(4),
        wait=wait_exponential(multiplier=1, min=2, max=6),
        retry=retry_if_exception_type(
            (
                httpx.HTTPStatusError,
                httpx.NetworkError,
                httpx.TimeoutException,
                openai.error.OpenAIError,
            )
        ),
    )
    async def _handle_response(
        self,
        index: str,
        prompt: str,
        tokens: int,
        raw_files: List[Tuple[str, str]] = None,
    ) -> Tuple[str, str]:
        """Handles OpenAI API response and returns the generated text."""
        token_cnt = count_tokens(prompt, self.encoder)
        if token_cnt > self.tokens_max:
            self.logger.debug(
                f"Truncating '{index}' prompt: {token_cnt} > {self.tokens_max}"
            )
            prompt = truncate_tokens(self.encoder, prompt, tokens)

        try:
            async with self.rate_limit_semaphore:
                response = await self.http_client.post(
                    self.endpoint,
                    headers={"Authorization": f"Bearer {openai.api_key}"},
                    json={
                        "messages": [
                            {
                                "role": "system",
                                "content": self.content,
                            },
                            {"role": "user", "content": prompt},
                        ],
                        "model": self.model,
                        "temperature": self.temperature,
                        "max_tokens": tokens,
                    },
                )
                response.raise_for_status()
                response_json = response.json()
                response_text = response_json["choices"][0]["message"][
                    "content"
                ]
                formatted_text = format_response(index, response_text)
                self.logger.info(f"Response for '{index}':\n{formatted_text}")
                return index, formatted_text

        except (
            httpx.HTTPStatusError,
            httpx.NetworkError,
            httpx.TimeoutException,
            openai.error.OpenAIError,
        ) as exc:
            self.logger.error(f"Error generating response for {index}: {exc}")
            return index, self.config.md.default



"""Model factory to return the appropriate LLM implementation module."""

from readmeai.config.settings import AppConfig
from readmeai.exceptions import UnsupportedServiceError
from readmeai.llms.offline import OfflineModeHandler
from readmeai.llms.openai import OpenAIHandler
from readmeai.llms.vertex import VertexAIHandler

MODEL_REGISTRY = {
    "offline": OfflineModeHandler,
    "openai": OpenAIHandler,
    "vertex": VertexAIHandler,
}


def model_handler(config: AppConfig) -> MODEL_REGISTRY:
    """Returns the appropriate LLM implementation."""
    llm_service = config.llm.api.lower()
    llm_module = MODEL_REGISTRY.get(llm_service)
    if not llm_module:
        raise UnsupportedServiceError(
            llm_service, f"LLM API service settings: {config.llm.api}"
        )
    return llm_module(config)



"""Builds each section of the README Markdown file."""

__package__ = "readmeai"

from pathlib import Path
from typing import List

from readmeai.config.enums import BadgeOptions, GitService
from readmeai.config.settings import AppConfig, ConfigHelper
from readmeai.markdown import badges, tables, tree
from readmeai.markdown.quickstart import setup_guide
from readmeai.markdown.utils import remove_emojis


class ReadmeBuilder:
    """Builds each section of the README Markdown file."""

    def __init__(
        self,
        conf: AppConfig,
        helper: ConfigHelper,
        dependencies: List[str],
        summaries: tuple,
        temp_dir: str,
    ):
        """Initializes the ReadmeBuilder class."""
        self.conf = conf
        self.helper = helper
        self.deps = dependencies
        self.summaries = summaries
        self.temp_dir = Path(temp_dir)
        self.md = self.conf.md
        self.full_name = self.conf.git.full_name
        self.host = self.conf.git.host
        self.host_domain = self.conf.git.host_domain
        self.repo_name = self.conf.git.name
        self.repo_url = self.conf.git.repository
        if self.host_domain == GitService.LOCAL:
            self.repo_url = f"../{self.repo_name}"

    @property
    def md_header(self) -> str:
        """Generates the README header section."""
        if BadgeOptions.SKILLS.value not in self.md.badge_style:
            md_shields, md_badges = badges.shields_icons(
                self.conf, self.deps, self.full_name, self.host
            )
        else:
            md_shields = (
                "<!-- Shields.io badges not used with skill icons. -->"
            )
            md_badges = badges.skill_icons(self.conf, self.deps)

        return self.md.header.format(
            align=self.md.align,
            image=self.md.image,
            repo_name=self.repo_name.upper(),
            slogan=self.md.slogan,
            badges_shields=md_shields,
            badges=md_badges,
        )

    @property
    def md_summaries(self) -> str:
        """Generates the README code summaries section."""
        summaries = tables.format_code_summaries(
            self.md.default,
            self.summaries,
        )
        md_summaries = tables.generate_markdown_tables(
            self.md.modules_widget,
            summaries,
            self.full_name,
            self.repo_url,
        )
        return md_summaries

    @property
    def md_tree(self) -> str:
        """Generates the README directory tree structure."""
        md_tree = tree.TreeGenerator(
            repo_name=self.repo_name,
            root_dir=self.temp_dir,
            repo_url=self.repo_url,
            max_depth=self.md.tree_depth,
        ).tree()
        return self.md.tree.format(md_tree)

    @property
    def md_quickstart(self) -> str:
        """Generates the README Getting Started section."""
        project_setup = setup_guide(self.conf, self.helper, self.summaries)
        return self.md.quickstart.format(
            repo_name=self.repo_name,
            repo_url=self.repo_url,
            requirements=project_setup.requirements,
            install_command=project_setup.install_command,
            run_command=project_setup.run_command,
            test_command=project_setup.test_command,
        )

    @property
    def md_contributing(self) -> str:
        """Generates the README Contributing section."""
        return self.md.contribute.format(
            host=self.host,
            host_domain=self.host_domain,
            full_name=self.full_name,
            repo_name=self.repo_name,
            repo_url=self.repo_url,
        )

    def build(self) -> str:
        """Builds the README Markdown file."""
        md_contents = [
            self.md_header,
            self.md.toc.format(repo_name=self.repo_name),
            self.md.overview,
            self.md.features,
            self.md_tree,
            self.md.modules,
            self.md_summaries,
            self.md_quickstart,
            self.md_contributing,
        ]

        if self.conf.md.emojis is False:
            md_contents = remove_emojis(md_contents)

        return "\n".join(md_contents)


def build_readme_md(
    conf: AppConfig,
    helper: ConfigHelper,
    dependencies: list,
    summaries: tuple,
    temp_dir: str,
) -> None:
    """Builds the README Markdown file."""
    builder = ReadmeBuilder(conf, helper, dependencies, summaries, temp_dir)
    markdown_sections = builder.build()
    return markdown_sections



"""Utility methods for the readme-ai CLI application."""

import os
import re
from importlib import resources
from pathlib import Path
from typing import Any

from readmeai.config.enums import LLMOptions, SecretKeys
from readmeai.config.settings import AppConfig, ConfigHelper
from readmeai.core.logger import Logger

logger = Logger(__name__)


def format_md_table(text: str) -> str:
    """
    Pattern to match a Markdown table. Looks for a
    header row with at least two columns. followed by
    a separator row, and then one or more data rows.
    """
    pattern = r"(\|.*\|.*\n\|[-: ]+\|[-: ]+\|.*\n(?:\|.*\|.*\n)*)"
    match = re.search(pattern, text, re.MULTILINE)
    return match.group(0) if match else ""


def format_md_text(text: str) -> str:
    """Format and clean generated text from the LLM."""
    # Remove single and double quotes around any text
    text = re.sub(r"(?<!\w)['\"](.*?)['\"](?!\w)", r"\1", text)

    # Remove newlines and tabs
    text = text.replace("\n", "").replace("\t", "")

    # Remove non-letter characters from the beginning of the string
    text = re.sub(r"^[^a-zA-Z]*", "", text)

    # Remove extra white space around punctuation except for '('
    text = re.sub(r"\s*([)'.!,?;:])(?!\.\s*\w)", r"\1", text)

    # Remove extra white space before opening parentheses
    text = re.sub(r"(\()\s*", r"\1", text)

    # Replace multiple consecutive spaces with a single space
    text = re.sub(r" +", " ", text)

    # Remove extra white space around hyphens
    text = re.sub(r"\s*-\s*", "-", text)

    return text.strip()


def format_response(prompt_type: str, response_text: str) -> str:
    """Post-processes the response from the LLM."""
    return (
        format_md_text(response_text)
        if prompt_type != "features"
        else format_md_table(response_text)
    )


def get_resource_path(package: str, resource_name: str) -> Path:
    """Use importlib.resources or pkg_resources to get resource path.
    Python 3.9+ uses importlib.resources, older versions use pkg_resources.

    Parameters
    ----------
    package
        Python package name.
    resource_name
        Name of the resource to get the path for.

    Returns
    -------
    Path
        The path to the resource file.
    """
    try:
        resource_path = resources.files(package) / resource_name

    except (TypeError, FileNotFoundError):
        import pkg_resources

        resource_path = pkg_resources.resource_filename(package, resource_name)

    return resource_path


def is_file_ignored(conf_helper: ConfigHelper, file_path: Path) -> bool:
    """Check if a file should be ignored and not passed to the LLM API.
    Uses a default list of files - 'readmeai/settings/ignore_files.toml'.

    Parameters
    ----------
    conf_helper
        Configuration helper instance with the ignore files list.
    file_path
        The path to the file to check.

    Returns
    -------
        True if the file should be ignored, False otherwise.
    """
    ignore_files = conf_helper.ignore_files

    if (
        (file_path.name in ignore_files["files"])
        or (file_path.suffix.lstrip(".") in ignore_files["extensions"])
        or any(
            directory in file_path.parts
            for directory in ignore_files["directories"]
        )
    ):
        return True

    return False


def scan_environment(keys: Any) -> bool:
    """Check if both keys in the tuple exist in the os.environ."""
    logger.debug(f"Scanning environment for keys: {keys}")
    return all(key in os.environ for key in keys)


def setup_environment(config: AppConfig, engine: str) -> None:
    """Set LLM environment variables based on the specified LLM service."""
    openai_keys = scan_environment((SecretKeys.OPENAI_API_KEY.value,))
    vertex_keys = scan_environment(
        (
            SecretKeys.VERTEXAI_LOCATION.value,
            SecretKeys.VERTEXAI_PROJECT.value,
        )
    )

    if engine is None:
        logger.info(
            "LLM engine not specified. Checking environment variables."
        )
        if openai_keys is True:
            logger.info("Using OpenAI API settings found in environment.")
            config.llm.api = LLMOptions.OPENAI.name
        elif vertex_keys is True:
            logger.info("Using GCP Vertex AI settings found in environment.")
            config.llm.api = LLMOptions.VERTEX.name
            config.llm.model = "gemini-pro"
        else:
            _enable_offline_mode(
                config, "No LLM API settings exist in environment!"
            )
    else:
        if engine == LLMOptions.OPENAI.name:
            logger.info("OpenAI LLM API service specified.")
            if openai_keys is False:
                _enable_offline_mode(
                    config, "OpenAI API settings NOT FOUND in environment!"
                )
        elif engine == LLMOptions.VERTEX.name:
            if vertex_keys is True:
                logger.info("Vertex AI environment variables found.")
                config.llm.model = "gemini-pro"
            else:
                _enable_offline_mode(
                    config, "Vertex AI settings NOT FOUND in environment!"
                )
        else:
            _enable_offline_mode(config, "Invalid LLM API service specified!")


def _enable_offline_mode(config: AppConfig, message: str) -> None:
    """Set the LLM service to offline mode."""
    config.llm.api = LLMOptions.OFFLINE.name
    config.llm.offline = True
    logger.warning(message)
    logger.warning("Generating README file in offline mode...")
    return



"""Methods to validate command-line arguments and settings."""

import re
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse, urlsplit

from readmeai.config.enums import GitService
from readmeai.core.logger import Logger
from readmeai.exceptions import GitValidationError

logger = Logger(__name__)


class GitValidator:
    """Validator class for GitSettings."""

    @classmethod
    def validate_repository(cls, value: Union[str, Path]) -> Union[str, Path]:
        """Validate the repository URL or path."""
        value_str = str(value)
        if (
            any(service.value in value_str for service in GitService) is False
            and not Path(value_str).is_dir()
        ):
            raise GitValidationError(value)

        if isinstance(value, str):
            path = Path(value)
            if path.is_dir():
                return value
            try:
                parsed_url = urlparse(value)
                if parsed_url.scheme in ["http", "https"] and any(
                    service in parsed_url.netloc for service in GitService
                ):
                    return value

            except Exception as exc:
                logger.error(f"Error validating repository: {exc}")
                raise GitValidationError(value) from exc

        elif isinstance(value, Path) and value.is_dir():
            return value

        raise GitValidationError(value)

    @classmethod
    def validate_full_name(cls, value: Optional[str], values: dict) -> str:
        """Validator for getting the full name of the repository."""
        url_or_path = values.get("repository")

        path = (
            url_or_path if isinstance(url_or_path, Path) else Path(url_or_path)
        )
        if path.exists():
            return str(path.name)

        patterns = {
            GitService.GITHUB: r"https?://github.com/([^/]+)/([^/]+)",
            GitService.GITLAB: r"https?://gitlab.com/([^/]+)/([^/]+)",
            GitService.BITBUCKET: r"https?://bitbucket.org/([^/]+)/([^/]+)",
        }

        for _, pattern in patterns.items():
            match = re.match(pattern, url_or_path)
            if match:
                user_name, repo_name = match.groups()
                return f"{user_name}/{repo_name}"

        raise GitValidationError(url_or_path)

    @classmethod
    def set_host(cls, value: Optional[str], values: dict) -> str:
        """Sets the Git service host from the repository provided."""
        repo = values.get("repository")
        if isinstance(repo, Path) or (
            isinstance(repo, str) and Path(repo).is_dir()
        ):
            return GitService.LOCAL

        parsed_url = urlparse(str(repo))
        for service in GitService:
            if service in parsed_url.netloc:
                return service.split(".")[0]

        return GitService.LOCAL

    @classmethod
    def set_host_domain(cls, value: Optional[str], values: dict) -> str:
        repo = values.get("repository")
        if isinstance(repo, Path) or (
            isinstance(repo, str) and Path(repo).is_dir()
        ):
            return GitService.LOCAL

        parsed_url = urlparse(str(repo))
        for service in GitService:
            if service in parsed_url.netloc:
                return service
        return GitService.LOCAL

    @classmethod
    def set_name(cls, value: Optional[str], values: dict) -> str:
        """Sets the repository name from the repository provided."""
        repo = values.get("repository")
        if isinstance(repo, Path):
            return repo.name
        elif isinstance(repo, str):
            parsed_url = urlsplit(repo)
            name = parsed_url.path.split("/")[-1]
            return name.removesuffix(".git")
        return "n/a"
